import subprocess
import requests
import os
import json
import PIL

import sqlite3
from datetime import datetime
from typing import List, Dict, Any, Optional, Union, Generator


from jinja2 import Environment, FileSystemLoader, Template, Undefined

import pandas as pd
import numpy as np
from sqlalchemy import create_engine

from npcpy.npc_sysenv import (
    get_system_message,
    get_available_models,
    get_model_and_provider,
    render_markdown,
    lookup_provider,
    NPCSH_CHAT_PROVIDER,
    NPCSH_CHAT_MODEL,
    NPCSH_API_URL,
    NPCSH_EMBEDDING_MODEL,
    NPCSH_EMBEDDING_PROVIDER,
    NPCSH_DEFAULT_MODE,
    NPCSH_REASONING_MODEL,
    NPCSH_REASONING_PROVIDER,
    NPCSH_IMAGE_GEN_MODEL,
    NPCSH_IMAGE_GEN_PROVIDER,
    NPCSH_VIDEO_GEN_MODEL,
    NPCSH_VIDEO_GEN_PROVIDER,
    NPCSH_VISION_MODEL,
    NPCSH_VISION_PROVIDER,
    available_reasoning_models,
    available_chat_models,
)
from npcpy.gen.response import get_litellm_response
from npcpy.gen.image_gen import generate_image_litellm
from npcpy.gen.video_gen import generate_video_diffusers

def generate_image(
    prompt: str,
    model: str = NPCSH_IMAGE_GEN_MODEL,
    provider: str = NPCSH_IMAGE_GEN_PROVIDER,
    filename: str = None,
    npc: Any = None,
    height: int = 256,
    width: int = 256,
):
    """This function generates an image using the specified provider and model.
    Args:
        prompt (str): The prompt for generating the image.
    Keyword Args:
        model (str): The model to use for generating the image.
        provider (str): The provider to use for generating the image.
        filename (str): The filename to save the image to.
        npc (Any): The NPC object.
    Returns:
        str: The filename of the saved image.
    """
    if model is not None and provider is not None:
        pass
    elif model is not None and provider is None:
        provider = lookup_provider(model)
    elif npc is not None:
        if npc.provider is not None:
            provider = npc.provider
        if npc.model is not None:
            model = npc.model
        if npc.api_url is not None:
            api_url = npc.api_url
    if filename is None:
        # Generate a filename based on the prompt and the date time
        os.makedirs(os.path.expanduser("~/.npcsh/images/"), exist_ok=True)
        filename = (
            os.path.expanduser("~/.npcsh/images/")
            + f"image_{datetime.now().strftime('%Y%m%d_%H%M%S')}.png"
        )
    image = generate_image_litellm(
        prompt=prompt,
        model=model,
        provider=provider,
        height=height,
        width=width,
    )
    print(image)
    # save image
    # check if image is a PIL image
    if isinstance(image, PIL.Image.Image):
        image.save(filename)
        return filename

    else:
        try:
            # image is at a private url
            response = requests.get(image.data[0].url)
            with open(filename, "wb") as file:
                file.write(response.content)
            from PIL import Image

            img = Image.open(filename)
            img.show()
            # console = Console()
            # console.print(Image.from_path(filename))
            return filename

        except AttributeError as e:
            print(f"Error saving image: {e}")


def generate_video(
    prompt,
    model: str = NPCSH_VIDEO_GEN_MODEL,
    provider: str = NPCSH_VIDEO_GEN_PROVIDER,
    npc: Any = None,
    device: str = "cpu",
    output_path="",
    num_inference_steps=10,
    num_frames=10,
    height=256,
    width=256,
    messages: list = None,
):
    """
    Function Description:
        This function generates a video using the Stable Diffusion API.
    Args:
        prompt (str): The prompt for generating the video.
        model_id (str): The Hugging Face model ID to use for Stable Diffusion.
        device (str): The device to run the model on ('cpu' or 'cuda').
    Returns:
        PIL.Image: The generated image.
    """
    output_path = generate_video_diffusers(
        prompt,
        model,
        npc=npc,
        device=device,
        output_path=output_path,
        num_inference_steps=num_inference_steps,
        num_frames=num_frames,
        height=height,
        width=width,
    )
    if provider == "diffusers":
        return {"output": "output path at " + output_path, "messages": messages}





def get_llm_response(
    prompt: str,
    provider: str = None,
    model: str=None,
    images: List[Dict[str, str]] = None,
    npc: Any = None,
    team: Any = None,
    messages: List[Dict[str, str]] = None,
    api_url: str = NPCSH_API_URL,
    api_key: str = None,
    context=None,    
    stream: bool = False,
    **kwargs,
):
    """This function generates a response using the specified provider and model.
    Args:
        prompt (str): The prompt for generating the response.
    Keyword Args:
        provider (str): The provider to use for generating the response.
        model (str): The model to use for generating the response.
        images (List[Dict[str, str]]): The list of images.
        npc (Any): The NPC object.
        messages (List[Dict[str, str]]): The list of messages.
        api_url (str): The URL of the API endpoint.
    Returns:
        Any: The response generated by the specified provider and model.
    """
    # print(provider, model)

    response = get_litellm_response(
        prompt= prompt,
        messages = messages,
        model=model,
        provider=provider,
        npc=npc,
        team=team,
        api_url=api_url,
        api_key=api_key,
        images=images,        
        context=context,
        stream=stream,
        **kwargs,
    )
    return response




def decide_plan(
    command: str,
    possible_actions: Dict[str, str],
    responder = None,
    messages: Optional[List[Dict[str, str]]] = None, 
    model: Optional[str] = None,
    provider: Optional[str] = None,
    api_url: Optional[str] = None,
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Gets a plan by asking LLM to choose from possible_actions.
    Prompt includes only command, actions, and potential delegation targets (if responder is Team).
    Relies on get_llm_response to handle all other context via the passed npc object.

    Args:
        command: User input.
        possible_actions: Dictionary of valid action names and descriptions.
        responder: The NPC or Team object, or None.
        messages: Conversation history (passed to get_llm_response).
        model, provider, api_url, api_key: LLM configuration overrides.

    Returns:
        Plan dictionary: {"chosen_action":..., "parameters":..., "explanation":..., "error":...}.
    """
    if not possible_actions:
        return {"error": "No possible actions provided."}


    prompt = f"User Request: \"{command}\"\n\n"
    prompt += "## Possible Actions:\nChoose ONE action:\n"
    for name, desc in possible_actions.items():
        prompt += f"- {name}: {desc}\n"

    # Add delegation targets ONLY if responder is a Team    
    if hasattr(responder, 'npcs'):
        foreman = decision_npc_obj # Already fetched
        entities = []
        for name, npc_obj in responder.npcs.items():
             if npc_obj != foreman: # Don't list the foreman itself
                 entities.append(f"- {name} (NPC)")
        for name, team_obj in responder.sub_teams.items():
             entities.append(f"- {name} (Team)")
        if entities:
            prompt += "\n### Potential Delegation Targets:\n" + "\n".join(entities) + "\n"

    # Instructions for LLM
    prompt += """
# Instructions:
1. Select the single most appropriate action from 'Possible Actions'.
2. Determine parameters identifying the action's target (e.g., 'tool_name', 'target_entity'). DO NOT determine tool arguments.
3. Explain reasoning.

# Required Output Format (JSON Only, NO MARKDOWN):
{
  "chosen_action": "action_name",
  "parameters": { /* "tool_name": "...", "target_entity": "..." */ },
  "explanation": "Your reasoning."
}
"""

    # --- 3. Call LLM ---
    # Pass decision_npc_obj - get_llm_response handles its context (directive etc.)
    llm_response = get_llm_response(
        prompt=prompt,
        model=effective_model,
        provider=effective_provider,
        api_url=effective_api_url,
        api_key=effective_api_key,
        npc=decision_npc_obj, # Pass the relevant NPC/Foreman object or None
        format="json",
        messages=messages # Pass history if get_llm_response uses it
    )

    # --- 4. Parse and Validate ---
    plan = {"error": None}
    if "error" in llm_response or "Error" in llm_response:
        plan["error"] = f"LLM Error: {llm_response.get('error', llm_response.get('Error'))}"
        print(plan["error"])
        return plan

    response_content = llm_response.get("response", {})
    parsed_json = None
    try:
        if isinstance(response_content, str): parsed_json = json.loads(response_content.strip())
        elif isinstance(response_content, dict): parsed_json = response_content
        else: raise TypeError("Response is not str or dict")

        if not all(k in parsed_json for k in ["chosen_action", "parameters", "explanation"]) or \
           not isinstance(parsed_json.get("parameters"), dict) or \
           not isinstance(parsed_json.get("chosen_action"), str) or \
           not parsed_json.get("chosen_action"):
            raise ValueError("LLM plan has invalid structure or missing keys.")

        chosen_action = parsed_json["chosen_action"]
        if chosen_action not in possible_actions:
            raise ValueError(f"LLM chose an invalid action '{chosen_action}'. Valid: {list(possible_actions.keys())}")

        plan["chosen_action"] = chosen_action
        plan["parameters"] = parsed_json.get("parameters", {})
        plan["explanation"] = parsed_json.get("explanation", "")

    except (json.JSONDecodeError, TypeError, ValueError) as e:
        plan["error"] = f"LLM plan processing error: {e}. Response: {str(response_content)[:200]}..."
        print(plan["error"])
        return plan


    print(f"Plan Determined: Action='{plan['chosen_action']}', Params='{plan['parameters']}'")
    return plan

def execute_llm_command(
    command: str,
    model: Optional[str] = None,
    provider: Optional[str] = None,
    api_url: str = NPCSH_API_URL,
    api_key: str = None,
    npc: Optional[Any] = None,
    messages: Optional[List[Dict[str, str]]] = None,
    stream=False,
    context=None,
) -> str:
    """This function executes an LLM command.
    Args:
        command (str): The command to execute.

    Keyword Args:
        model (Optional[str]): The model to use for executing the command.
        provider (Optional[str]): The provider to use for executing the command.
        npc (Optional[Any]): The NPC object.
        messages (Optional[List[Dict[str, str]]): The list of messages.
    Returns:
        str: The result of the LLM command.
    """

    max_attempts = 5
    attempt = 0
    subcommands = []
    npc_name = npc.name if npc else "sibiji"
    location = os.getcwd()
    print(f"{npc_name} generating command")
    # Create context from retrieved documents
    context = ""
    while attempt < max_attempts:
        prompt = f"""
        A user submitted this query: {command}.
        You need to generate a bash command that will accomplish the user's intent.
        Respond ONLY with the command that should be executed.
        in the json key "bash_command".
        You must reply with valid json and nothing else. Do not include markdown formatting
        """
        if len(context) > 0:
            prompt += f"""
            What follows is the context of the text files in the user's directory that are potentially relevant to their request
            Use these to help inform your decision.
            {context}
            """
        if len(messages) > 0:
            prompt += f"""
            The following messages have been exchanged between the user and the assistant:
            {messages}
            """

        response = get_llm_response(
            prompt,
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            messages=[],
            npc=npc,
            format="json",
            context=context,
        )

        llm_response = response.get("response", {})
        # messages.append({"role": "assistant", "content": llm_response})
        # print(f"LLM response type: {type(llm_response)}")
        # print(f"LLM response: {llm_response}")

        try:
            if isinstance(llm_response, str):
                llm_response = json.loads(llm_response)

            if isinstance(llm_response, dict) and "bash_command" in llm_response:
                bash_command = llm_response["bash_command"]
            else:
                raise ValueError("Invalid response format from LLM")
        except (json.JSONDecodeError, ValueError) as e:
            print(f"Error parsing LLM response: {e}")
            attempt += 1
            continue

        print(f"LLM suggests the following bash command: {bash_command}")
        subcommands.append(bash_command)

        try:
            print(f"Running command: {bash_command}")
            result = subprocess.run(
                bash_command, shell=True, text=True, capture_output=True, check=True
            )
            print(f"Command executed with output: {result.stdout}")

            prompt = f"""
                Here was the output of the result for the {command} inquiry
                which ran this bash command {bash_command}:

                {result.stdout}

                Provide a simple response to the user that explains to them
                what you did and how it accomplishes what they asked for.
                """
            if len(context) > 0:
                prompt += f"""
                What follows is the context of the text files in the user's directory that are potentially relevant to their request
                Use these to help inform how you respond.
                You must read the context and use it to provide the user with a more helpful answer related to their specific text data.

                CONTEXT:

                {context}
                """
            messages.append({"role": "user", "content": prompt})
            # print(messages, stream)
            response = get_llm_response(
                prompt,
                model=model,
                provider=provider,
                api_url=api_url,
                api_key=api_key,
                npc=npc,
                messages=messages,
                context=context,
            )
            output = response.get("response", "")
            return {"messages": messages, "output": output}
        except subprocess.CalledProcessError as e:
            print(f"Command failed with error:")
            print(e.stderr)

            error_prompt = f"""
            The command '{bash_command}' failed with the following error:
            {e.stderr}
            Please suggest a fix or an alternative command.
            Respond with a JSON object containing the key "bash_command" with the suggested command.
            Do not include any additional markdown formatting.

            """

            if len(context) > 0:
                error_prompt += f"""
                    What follows is the context of the text files in the user's directory that are potentially relevant to their request
                    Use these to help inform your decision.
                    {context}
                    """

            fix_suggestion = get_llm_response(
                error_prompt,
                model=model,
                provider=provider,
                npc=npc,
                api_url=api_url,
                api_key=api_key,
                format="json",
                messages=messages,
                context=context,
            )

            fix_suggestion_response = fix_suggestion.get("response", {})

            try:
                if isinstance(fix_suggestion_response, str):
                    fix_suggestion_response = json.loads(fix_suggestion_response)

                if (
                    isinstance(fix_suggestion_response, dict)
                    and "bash_command" in fix_suggestion_response
                ):
                    print(
                        f"LLM suggests fix: {fix_suggestion_response['bash_command']}"
                    )
                    command = fix_suggestion_response["bash_command"]
                else:
                    raise ValueError(
                        "Invalid response format from LLM for fix suggestion"
                    )
            except (json.JSONDecodeError, ValueError) as e:
                print(f"Error parsing LLM fix suggestion: {e}")

        attempt += 1

    return {
        "messages": messages,
        "output": "Max attempts reached. Unable to execute the command successfully.",
    }

def handle_tool_call(
    command: str,
    tool_name: str,
    model: str = NPCSH_CHAT_MODEL,
    provider: str = NPCSH_CHAT_PROVIDER,
    api_url: str = NPCSH_API_URL,
    api_key: str = None,
    messages: List[Dict[str, str]] = None,
    npc: Any = None,
    stream=False,
    n_attempts=3,
    attempt=0,
    context=None,
) -> Union[str, Dict[str, Any]]:
    """This function handles a tool call.
    Args:
        command (str): The command.
        tool_name (str): The tool name.
    Keyword Args:
        model (str): The model to use for handling the tool call.
        provider (str): The provider to use for handling the tool call.
        messages (List[Dict[str, str]]): The list of messages.
        npc (Any): The NPC object.
    Returns:
        Union[str, Dict[str, Any]]: The result of handling
        the tool call.

    """
    print(npc)
    print("handling tool call")
    print(command)
    if npc is None:
        return f"No tools are available. "
    else:
        if tool_name not in npc.tools_dict:
            print("not available")
            print(f"Tool '{tool_name}' not found in NPC's tools_dict.")
            return f"Tool '{tool_name}' not found."
        elif tool_name in npc.tools_dict:
            tool = npc.tools_dict[tool_name]
        print(f"Tool found: {tool.tool_name}")
        jinja_env = Environment(loader=FileSystemLoader("."), undefined=Undefined)

        prompt = f"""
        The user wants to use the tool '{tool_name}' with the following request:
        '{command}'
        Here is the tool file:
        ```
        {tool.to_dict()}
        ```

        Please determine the required inputs for the tool as a JSON object.
        
        
        They must be exactly as they are named in the tool.
        For example, if the tool has three inputs, you should respond with a list of three values that will pass for those args.
        
        Return only the JSON object without any markdown formatting.

        """

        if npc and hasattr(npc, "shared_context"):
            if npc.shared_context.get("dataframes"):
                context_info = "\nAvailable dataframes:\n"
                for df_name in npc.shared_context["dataframes"].keys():
                    context_info += f"- {df_name}\n"
                prompt += f"""Here is contextual info that may affect your choice: {context_info}
                """
        if context is not None:
            prompt += f"Here is some additional context: {context}"

        # print(prompt)

        # print(
        # print(prompt)
        response = get_llm_response(
            prompt,
            format="json",
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            npc=npc,
        )
        try:
            # Clean the response of markdown formatting
            response_text = response.get("response", "{}")
            if isinstance(response_text, str):
                response_text = (
                    response_text.replace("```json", "").replace("```", "").strip()
                )

            # Parse the cleaned response
            if isinstance(response_text, dict):
                input_values = response_text
            else:
                input_values = json.loads(response_text)
            # print(f"Extracted inputs: {input_values}")
        except json.JSONDecodeError as e:
            print(f"Error decoding input values: {e}. Raw response: {response}")
            return f"Error extracting inputs for tool '{tool_name}'"
        # Input validation (example):
        required_inputs = tool.inputs
        missing_inputs = []
        for inp in required_inputs:
            if not isinstance(inp, dict):
                # dicts contain the keywords so its fine if theyre missing from the inputs.
                if inp not in input_values or input_values[inp] == "":
                    missing_inputs.append(inp)
        if len(missing_inputs) > 0:
            # print(f"Missing required inputs for tool '{tool_name}': {missing_inputs}")
            if attempt < n_attempts:
                print(f"attempt {attempt+1} to generate inputs failed, trying again")
                print("missing inputs", missing_inputs)
                print("llm response", response)
                print("input values", input_values)
                return handle_tool_call(
                    command,
                    tool_name,
                    model=model,
                    provider=provider,
                    messages=messages,
                    npc=npc,
                    api_url=api_url,
                    api_key=api_key,
                    stream=stream,
                    attempt=attempt + 1,
                    n_attempts=n_attempts,
                )
            return {
                "output": f"Missing inputs for tool '{tool_name}': {missing_inputs}",
                "messages": messages,
            }

        # try:
        print("Executing tool with input values:", input_values)

        # try:
        tool_output = tool.execute(
            input_values,
            npc.tools_dict,
            jinja_env,
            command,
            model=model,
            provider=provider,
            npc=npc,
            stream=stream,
            messages=messages,
        )
        if not stream:
            if "Error" in tool_output:
                raise Exception(tool_output)
            # except Exception as e:
            # diagnose_problem = get_llm_response(
            ##    f"""a problem has occurred.
            #                                    Please provide a diagnosis of the problem and a suggested #fix.

            #                                    The tool call failed with this error:
            #                                    {e}
            #                                    Please return a json object containing two fields
            ##                                    -problem
            #                                    -suggested solution.
            #                                    do not include any additional markdown formatting or #leading json tags

            #                                    """,
            #    model=model,
            #    provider=provider,
            #    npc=npc,
            ##    api_url=api_url,
            #    api_ley=api_key,
            #    format="json",
            # )
            # print(e)
            # problem = diagnose_problem.get("response", {}).get("problem")
            # suggested_solution = diagnose_problem.get("response", {}).get(
            #    "suggested_solution"
            # )
            '''
            print(f"An error occurred while executing the tool: {e}")
            print(f"trying again, attempt {attempt+1}")
            if attempt < n_attempts:
                tool_output = handle_tool_call(
                    command,
                    tool_name,
                    model=model,
                    provider=provider,
                    messages=messages,
                    npc=npc,
                    api_url=api_url,
                    api_key=api_key,
                    stream=stream,
                    attempt=attempt + 1,
                    n_attempts=n_attempts,
                    context=f""" \n \n \n "tool failed: {e}  \n \n \n here was the previous attempt: {input_values}""",
                )
            else:
                user_input = input(
                    "the tool execution has failed after three tries, can you add more context to help or would you like to run again?"
                )
                return
            '''
        if stream:
            return tool_output
        # print(f"Tool output: {tool_output}")
        # render_markdown(str(tool_output))
        if messages is not None:  # Check if messages is not None
            messages.append({"role": "assistant", "content": tool_output})
        return {"messages": messages, "output": tool_output}
        # except Exception as e:
        #    print(f"Error executing tool {tool_name}: {e}")
        #    return f"Error executing tool {tool_name}: {e}"



def check_llm_command(
    command: str,
    model: str = None, 
    provider: str  = None, 
    api_url: str = NPCSH_API_URL,
    api_key: str = None,
    npc: Any = None,
    team: Any = None,
    messages: List[Dict[str, str]] = None,
    images: list = None,
    stream=False,
    context=None,
    synthesize=False,
    human_in_the_loop=False,
):
    """This function checks an LLM command.
    Args:
        command (str): The command to check.
    Keyword Args:
        model (str): The model to use for checking the command.
        provider (str): The provider to use for checking the command.
        npc (Any): The NPC object.
        n_docs (int): The number of documents.
    Returns:
        Any: The result of checking the LLM command.
    """

    ENTER_REASONING_FLOW = False
    if NPCSH_DEFAULT_MODE == "reasoning":
        ENTER_REASONING_FLOW = True
    if model in available_reasoning_models:
        print(
            """
Model provided is a reasoning model, defaulting to non reasoning model for
ReAct choices then will enter reasoning flow
            """
        )
        model = NPCSH_CHAT_MODEL
        provider = NPCSH_CHAT_PROVIDER
    if messages is None:
        messages = []

    prompt = f"""
    A user submitted this query: {command}

    Determine the nature of the user's request:

    1. Should a tool be invoked to fulfill the request?

    2. Is it a general question that requires an informative answer or a highly specific question that
        requires inforrmation on the web?

    3. Would this question be best answered by an alternative NPC?

    4. Is it a complex request that actually requires more than one
        tool to be called, perhaps in a sequence?
        Sequences should only be used for more than one consecutive tool call. Do not invoke sequences for single tool calls.


    """
    if human_in_the_loop:

        prompt+= f"""
            5. is there a high amount of ambiguity in the user's request?  If so, ask the user for more information.
            
            in your response, consider as well the following guidelines for whether to request input:
            
                Here are some examples of ambiguous and non-ambiguous requests:

                For exmaple,
                    "tell me a joke about my favorite city" is ambiguous because the user
                    did not specify the city. In this case, ask the user for more information.

                    "tell me a joke about the weather" is not ambiguous because the user
                    specified the topic of the joke. In this case, you can answer the question
                    without asking for more information.

                    "take a screenshot of my screen" is not ambiguous because the user
                    specified the action they want to take. In this case, you can carry out the action without asking for more information.

                    ambiguous: "whats happening tonight in my city" is ambiguous because the user 
                    did not specify the city. In this case, ask the user for more information.
                    not ambiguous: "whats happening tonight in new york" is not ambiguous because the user
                    specified the city. In this case, you can answer the question without asking for more information.
                        
                Please limit requests for input to only the most ambiguous requests to ensure the optimal user experience.
            
            """
        

    if npc is not None:
        if npc.shared_context:
            prompt += f"""
            Relevant shared context for the npc:
            {npc.shared_context}
            """

        if npc.tools_dict is None :
            prompt += "No tools available. Do not invoke tools."
        else:
            prompt += "Available tools: \n"
            tools_set = {}
            if npc.tools_dict is not None:
                for tool_name, tool in npc.tools_dict.items():
                    if tool_name not in tools_set:
                        tools_set[tool_name] = tool.description
            for tool_name, tool_description in tools_set.items():
                prompt += f"""
                            {tool_name} : {tool_description} \n

                            """
    if team is None:
        prompt += "No NPCs available for alternative answers."
    else:
        prompt += f"""
        Available NPCs for alternative answers:

            {team.npcs}
        """
        if team.context:
            prompt += f"""
            Relevant shared context for the team:
            {team.context}
            """


    action_space = ["invoke_tool", "answer_question", "pass_to_npc", "execute_sequence", ]
    if human_in_the_loop:
        action_space.append("request_input")
    prompt += f"""
    In considering how to answer this, consider:

    - Whether a tool should be used.


    Excluding time-sensitive phenomena or ones that require external data inputs /information,
    most general questions can be answered without any extra tools or agent passes.


    Only use tools or pass to other NPCs when it is obvious that the answer needs to be as up-to-date as possible. For example,
        a question about where mount everest is does not necessarily need to be answered by a tool call or an agent pass.
    Similarly, if a user asks to explain the plot of the aeneid, this can be answered without a tool call or agent pass.

    If a user were to ask for the current weather in tokyo or the current price of bitcoin or who the mayor of a city is,
        then a tool call or agent pass may be appropriate.

    Tools are valuable but their use should be limited and purposeful to
        ensure the best user experience.
        If a user asks you to search or to take a screenshot or to open a program or to write a program most likely it is
        appropriate to use a tool.
    Respond with a JSON object containing:
    - "action": one of {action_space}
    - "tool_name": : if action is "invoke_tool": the name of the tool to use.
                     else if action is "execute_sequence", a list of tool names to use.
    - "explanation": a brief explanation of why you chose this action.
    - "npc_name": (if action is "pass_to_npc") the name of the NPC to pass the question , else if action is "execute_sequence", a list of
                    npcs to pass the question to in order.



    Return only the JSON object. Do not include any additional text.

    The format of the JSON object is:
    {{
        "action": "invoke_tool" | "answer_question" | "pass_to_npc" | "execute_sequence" | "request_input",
        "tool_name": "<tool_name(s)_if_applicable>",
        "explanation": "<your_explanation>",
        "npc_name": "<npc_name(s)_if_applicable>"
    }}

    If you execute a sequence, ensure that you have a specified NPC for each tool use.
        question answering is not a tool use.
        "invoke_tool" should never be used in the list of tools when executing a sequence.
    Remember, do not include ANY ADDITIONAL MARKDOWN FORMATTING.
    There should be no leading ```json.
    
        """
    if context:
        prompt += f"""
        Additional relevant context from user:

        {context}

        """


    action_response = get_llm_response(
        prompt,
        model=model,
        provider=provider,
        api_url=api_url,
        api_key=api_key,
        npc=npc,
        format="json",
        messages=[],
        context=None,
    )
    if "Error" in action_response:
        print(f"LLM Error: {action_response['error']}")
        return action_response["error"]

    response_content = action_response.get("response", {})

    if isinstance(response_content, str):
        try:
            response_content_parsed = json.loads(response_content)
        except json.JSONDecodeError as e:
            print(
                f"Invalid JSON received from LLM: {e}. Response was: {response_content}"
            )
            return f"Error: Invalid JSON from LLM: {response_content}"
    else:
        response_content_parsed = response_content

    action = response_content_parsed.get("action")
    explanation = response_content_parsed.get("explanation")

    render_markdown(f"- Action chosen: {action}\n")

    # print(response_content)
    if response_content_parsed.get("tool_name"):
        print(f"tool name: {response_content_parsed.get('tool_name')}")
    if action == "execute_command":

        result = execute_llm_command(
            command,
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            messages=[],
            npc=npc,
            stream=stream,
        )

        output = result.get("response", "")
        messages = result.get("messages", messages)
        return {"messages": messages, "output": output}

    elif action == "invoke_tool":
        tool_name = response_content_parsed.get("tool_name")
        # print(npc)
        print(f"tool name: {tool_name}")
        result = handle_tool_call(
            command,
            tool_name,
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            messages=messages,
            npc=npc,
            stream=stream,
        )

        messages = result.get("messages", messages)
        output = result.get("response", "")
        return {"messages": messages, "output": output}

    elif action == "answer_question":

        result = get_llm_response(
            command,
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            messages=messages,
            npc=npc,
            stream=stream,
            images=images,
        )
        messages = result.get("messages", messages)
        output = result.get("response", "")
        return {"messages": messages, "output": output}
    elif action == "pass_to_npc":
        npc_to_pass = response_content_parsed.get("npc_name")
        npc_to_pass_obj = None
        print(npc_to_pass)
        agent_passes = []
        if team is not None:
            print(f"team npcs: {team.npcs}")
            match = team.npcs.get(npc_to_pass)
            if match is not None:
                npc_to_pass_obj = match
                print(type(npc_to_pass_obj))
                agent_passes.append(
                    npc.handle_agent_pass(
                        npc_to_pass_obj,
                        command,
                        messages=messages,
                    )
                )
        output = ""
        print(agent_passes)
        for agent_pass in agent_passes:
            output += str(agent_pass.get("response"))
        return {"messages": messages, "output": output}
    elif action == "request_input":
        explanation = response_content_parsed.get("explanation")

        request_input = handle_request_input(
            f"Explanation from check_llm_command:  {explanation} \n for the user input command: {command}",
            model=model,
            provider=provider,
        )
        # pass it back through with the request input added to the end of the messages
        # so that we can re-pass the result through the check_llm_command.

        messages.append(
            {
                "role": "assistant",
                "content": f"""its clear that extra input is required.
                                could you please provide it? Here is the reason:

                                {explanation},

                                and the prompt: {command}""",
            }
        )
        messages.append(
            {
                "role": "user",
                "content": command + " \n \n \n extra context: " + request_input,
            }
        )

        return check_llm_command(
            command + " \n \n \n extra context: " + request_input,
            model=model,
            provider=provider,
            api_url=api_url,
            api_key=api_key,
            npc=npc,
            messages=messages,
            stream=stream,
        )

    elif action == "execute_sequence":
        tool_names = response_content_parsed.get("tool_name")
        npc_names = response_content_parsed.get("npc_name")

        # print(npc_names)
        npcs = []
        # print(tool_names, npc_names)
        if isinstance(npc_names, list):
            if len(npc_names) == 0:
                # if no npcs are specified, just have the npc take care of it itself instead of trying to force it to generate npc names for sequences all the time

                npcs = [npc] * len(tool_names)
            if team is None:
                # try again and append that there are no agents to pass to
                print('')
            for npc_name in npc_names:
                for npc_obj in team.npcs:
                    if npc_name in npc_obj:
                        npcs.append(npc_obj[npc_name])
                        break
                if len(npcs) < len(tool_names):
                    npcs.append(npc)

        output = ""
        results_tool_calls = []
        if synthesize:
            # carry out fact extraction
            print("synthesize not yet implemented")

        if len(tool_names) > 0:
            for npc_obj, tool_name in zip(npcs, tool_names):
                result = handle_tool_call(
                    command,
                    tool_name,
                    model=model,
                    provider=provider,
                    api_url=api_url,
                    api_key=api_key,
                    messages=messages,
                    npc=npc_obj,
                    stream=stream,
                )
                # print(result)
                results_tool_calls.append(result)
                messages = result.get("messages", messages)
                output += result.get("response", "")
                # print(results_tool_calls)
        else:
            print('agent pass')
            for npc_obj in npcs:
                result = npc.handle_agent_pass(
                    npc_obj,
                    command,
                    messages=messages,
                    shared_context=npc.shared_context,
                )

                messages = result.get("messages", messages)
                results_tool_calls.append(result.get("response"))
                # print(messages[-1])
        # import pdb

        # pdb.set_trace()

        return {"messages": messages, "output": output}
    else:
        print("Error: Invalid action in LLM response")
        return "Error: Invalid action in LLM response"


def rehash_last_message(
    conversation_id: str,
    model: str,
    provider: str,
    npc: Any = None,
    stream: bool = False,
) -> dict:
    from npcpy.memory.command_history import CommandHistory    
    command_history = CommandHistory()
    last_message = command_history.get_last_conversation(conversation_id)
    if last_message is None:
        convo_id = command_history.get_most_recent_conversation_id()[0]
        last_message = command_history.get_last_conversation(convo_id)

    user_command = last_message[3]  # Assuming content is in the 4th column
    return check_llm_command(
        user_command,
        model=model,
        provider=provider,
        npc=npc,
        messages=None,
        stream=stream,
    )
